# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox
from functools import partial

import logging,re,os,warnings
import numpy as np
import pandas as pd
import bdata as bd
import weakref as wref

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout,on_entry_click
from bfit.backend.raise_window import raise_window
from bfit.backend.ConstrainedFunction import ConstrainedFunction as CstrFnGenerator

import bfit.backend.colors as colors

from bfit.fitting.fit_bdata import fit_list

# ========================================================================== #
class popup_fit_constraints(object):
    """
        Popup window for modelling the fit results with a function
        
        bfit
        fittab
        logger
        
        constr_text:        string, text defining constraints
        entry:              Text, text entry for user
        new_par:            dataframe, index: parnames, columns: p0,blo,bhi,res,err
        
        output_par_text     text, detected parameter names
        output_text         dict, keys: p0,blo,bhi,res,err, value: tkk.Text objects
       
        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text
       
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}
    
    # default parameter values on new parameter
    default_parvals = { 'p0':1,
                        'blo':-np.inf,
                        'bhi':np.inf,
                        'res':np.nan,
                        'err':np.nan}

    # ====================================================================== #
    def __init__(self,bfit,constr_text='',output_par_text='',output_text=''):
        self.bfit = bfit
        self.fittab = bfit.fit_files
        self.constr_text = constr_text
        self.output_par_text_val = output_par_text
        
        if not output_text:
            self.output_text_val = {}
        else:
            self.output_text_val = output_text
    
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Fit With Constrained Parameters')
        frame = ttk.Frame(self.win,relief='sunken',pad=5)
        left_frame = ttk.Frame(frame)
        right_frame = ttk.Frame(frame)

        # set icon
        try:
            img = PhotoImage(file=os.path.dirname(__file__)+'/../images/icon.gif')
            self.win.tk.call('wm', 'iconphoto', self.win._w, img)
        except Exception as err:
            print(err)

        # Key bindings
        self.win.bind('<Control-Key-Return>',self.do_fit)
        self.win.bind('<Control-Key-KP_Enter>',self.do_fit)

        # Keyword parameters
        key_param_frame = ttk.Frame(left_frame,relief='sunken',pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = CstrFnGenerator.keyvars
        
        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len,keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk),d) for k,d in zip(keys,descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame,text=s,justify=LEFT)
        
        # fit parameter names 
        fit_param_frame = ttk.Frame(left_frame,relief='sunken',pad=5)
        s = 'Reserved function parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(),
                                        self.fittab.n_component.get())
        
        s += '\n'.join([k for k in sorted(self.parnames)]) 
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame,text=s,justify=LEFT)

        # module names 
        module_frame = ttk.Frame(left_frame,relief='sunken',pad=5)
        s = 'Reserved module names:\n\n'
        
        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len,keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk),d) for k,d in zip(keys,descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame,text=s,justify=LEFT)
        
        # Text entry
        entry_frame = ttk.Frame(right_frame,relief='sunken',pad=5)
        entry_label = ttk.Label(entry_frame,justify=LEFT,
                                text='Enter one constraint equation per line.'+\
                                     '\nNon-reserved words are shared variables.'+\
                                     '\nEx: "1_T1 = a*np.exp(b*BIAS**0.5)+c"'+\
                                     '\nNote: Shared and fixed flags from main window ignored.')
        self.entry = Text(entry_frame,width=60,height=13,state='normal')
        self.entry.bind('<KeyRelease>',self.get_input)
        scrollb = Scrollbar(entry_frame, command=self.entry.yview)
        self.entry['yscrollcommand'] = scrollb.set
        
        # Insert default text
        self.entry.insert('1.0',self.constr_text.strip())
        
        # text for output
        output_frame = ttk.Frame(right_frame,relief='sunken',pad=5)
        output_head1_label = ttk.Label(output_frame,text='Par Name')
        output_head2_label = ttk.Label(output_frame,text='p0')
        output_head3_label = ttk.Label(output_frame,text='Bounds')
        output_head4_label = ttk.Label(output_frame,text='Result')
        output_head5_label = ttk.Label(output_frame,text='Error')
        self.output_par_text = Text(output_frame,width=8,height=8)
        self.output_text = {k:Text(output_frame,width=8,height=8,wrap='none')\
                            for k in ('p0','blo','bhi','res','err')}
        
        # default starter strings
        if self.output_par_text_val: 
            self.output_par_text.insert('1.0',self.output_par_text_val)
        self.output_par_text.config(state='disabled')
    
        if self.output_text_val: 
            for k in self.output_text_val:
                self.output_text[k].insert('1.0',self.output_text_val[k])

        # disable results
        for k in ('res','err'):
            self.output_text[k].config(state='disabled',width=12)

        # key bindings and scrollbar
        scrollb_out = Scrollbar(output_frame, command=self.yview)
        self.output_par_text['yscrollcommand'] = scrollb_out.set
        for k in self.output_text:
            self.output_text[k].bind('<KeyRelease>',self.get_result_input)
            self.output_text[k]['yscrollcommand'] = scrollb_out.set
                
        c = 0; r = 0;
        output_head1_label.grid(column=c,row=r);        c+=1;
        output_head2_label.grid(column=c,row=r);        c+=1;
        output_head3_label.grid(column=c,row=r,
                                columnspan=2);          c+=2;
        output_head4_label.grid(column=c,row=r);        c+=1;
        output_head5_label.grid(column=c,row=r);        c+=1;
        
        c = 0; r += 1;
        self.output_par_text.grid(column=c,row=r,sticky=N); c+=1;
        for k in ('p0','blo','bhi','res','err'):
            self.output_text[k].grid(column=c,row=r,sticky=N); c+=1;
        scrollb_out.grid(row=r, column=c, sticky='nsew')
        
        # fitting button 
        fit_button = ttk.Button(right_frame,text='Fit',command=self.do_fit)
        
        # gridding
        key_param_label.grid(column=0,row=0)
        fit_param_label.grid(column=0,row=0)
        modules_label.grid(column=0,row=0)
        entry_label.grid(column=0,row=0,sticky=W)
        self.entry.grid(column=0,row=1)
        scrollb.grid(row=1, column=1, sticky='nsew')
        
        # grid to frame
        frame.grid(column=0,row=0)
        left_frame.grid(column=0,row=0,sticky=(N,S))
        right_frame.grid(column=1,row=0,sticky=(N,S))
        
        key_param_frame.grid(column=0,row=0,rowspan=1,sticky=(E,W),padx=1,pady=1)
        module_frame.grid(column=0,row=1,sticky=(E,W),padx=1,pady=1,rowspan=2)
        fit_param_frame.grid(column=0,row=3,sticky=(E,W,N,S),padx=1,pady=1)
        
        entry_frame.grid(column=0,row=0,sticky=(N,E,W),padx=1,pady=1)
        output_frame.grid(column=0,row=1,sticky=(N,E,W,S),padx=1,pady=1)
        fit_button.grid(column=0,row=2,sticky=(N,E,W),padx=1,pady=1)
        
        # initialize 
        self.new_par = pd.DataFrame(columns=['name','p0','blo','bhi','res','err']) 
        
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()
    
    # ====================================================================== #
    def do_fit(self,*args):
        """
            Set up the fit functions and do the fit. Then map the outputs to the
            proper displays. 
        """
        
        # parse text
        self.do_parse()
        
        # clean input
        text = self.constr_text.split('\n')
        text = [t.strip() for t in text if '=' in t]
        
        # check for no input
        if not text:    return
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # check that the defined variables all match function inputs
        for d in defined: 
            if d not in self.parnames:
                errmsg = 'Definition for "%s" invalid. ' % d+\
                         'Must only define function inputs. '
                messagebox.showerror("Error",errmsg)
                raise RuntimeError(errmsg)
        
        # make shared parameters for the rest of the parameters
        allpar = self.new_par['name'].tolist()
        alldef = defined[:]     # all parameter names in order
        sharelist = [True]*len(allpar)
        
        for n in sorted(self.parnames):
            if n not in defined:
                eqn.append(n)
                alldef.append(n)
                allpar.append(n)
                sharelist.append(False)
                        
        # replace 1_T1 with lambda1
        for i,_ in enumerate(allpar):
            if '1_T1' in allpar[i]:
                allpar[i] = allpar[i].replace('1_T1','lambda1')
        
        for i,_ in enumerate(eqn):
            while '1_T1' in eqn[i]:
                eqn[i] = eqn[i].replace('1_T1','lambda1')
                
        # make constrained functions
        cgen= CstrFnGenerator(self.bfit,alldef,eqn,allpar,self.parnames)
        
        # get the functions and initial parameters
        fit_files = self.bfit.fit_files
        fetch_files = self.bfit.fetch_files
        fitfns = []
        par = []
        rebin = []
        omit = []
        fnptrs = []
        constr_fns = []
        
        keylist = sorted(fit_files.fit_lines.keys())
        for k in keylist:
            line = fetch_files.data_lines[k]
            data = line.bdfit
            
            # get pulse length
            pulse_len = -1
            try:
                pulse_len = data.bd.get_pulse_s()
            except KeyError:
                pass
            
            # get function
            fn = fit_files.fitter.get_fn(fn_name=fit_files.fit_function_title.get(),
                                         ncomp=fit_files.n_component.get(),
                                         pulse_len=pulse_len,
                                         lifetime=bd.life[fit_files.probe_label['text']])
            
            genf,genc = cgen(data=data,fn=fn)
            fitfns.append(genf)
            fnptrs.append(fn)
            constr_fns.append(genc)
            
            # get initial parameters
            par.append(data.fitpar)
            
            # get rebin
            rebin.append(data.rebin.get())
            
            # get bin omission
            omit.append(data.omit.get())
        
        # clean up omit strings
        for i,om in enumerate(omit):
            if om == fetch_files.bin_remove_starter_line:
                omit[i] = ''
        
        # set up p0, bounds
        p0 = self.new_par['p0'].values
        blo = self.new_par['blo'].values
        bhi = self.new_par['bhi'].values
        
        p0 = [[p]*len(keylist) for p in p0]
        blo = [[p]*len(keylist) for p in blo]
        bhi = [[p]*len(keylist) for p in bhi]
                
        for n in sorted(self.parnames):
            if n not in defined:
                p0.append( [p['p0' ][n] for p in par])
                blo.append([p['blo'][n] for p in par])
                bhi.append([p['bhi'][n] for p in par])
        
        p0 = np.array(p0).T
        blo = np.array(blo).T
        bhi = np.array(bhi).T
        
        # set up fitter inputs
        runs = [int(k.split('.')[1]) for k in keylist]
        years = [int(k.split('.')[0]) for k in keylist]
        npar = len(sharelist)
        bounds = [[l,h] for l,h in zip(blo,bhi)]
        
        # do the fit and kill fitting window
        fit_status_window = fit_files.make_fit_status_window()
        try:
            par,cov,chi,gchi = fit_list(runs=runs,
                                        years=years,
                                        fnlist=fitfns,
                                        sharelist=sharelist,
                                        npar=npar,
                                        p0=p0,
                                        bounds=bounds,
                                        asym_mode='c',
                                        rebin=rebin,
                                        omit=omit,
                                        xlims=None,
                                        hist_select=self.bfit.hist_select)
        except Exception as errmsg:
            self.logger.exception('Fitting error')
            messagebox.showerror("Error",str(errmsg))
            raise errmsg from None
        finally:
            fit_status_window.destroy()
            del fit_status_window
        std = np.array(list(map(np.diag,cov)))**0.5
        
        # display output for global parameters
        for i in self.new_par.index:
            self.new_par.loc[i,'res'] = par[0][i]
            self.new_par.loc[i,'err'] = std[0][i]
        self.set_par_text()
        
        # calculate original parameter equivalents
        for i,k in enumerate(keylist):
            data = fetch_files.data_lines[k].bdfit
            
            # calculate
            old_par = [cfn(*par[i]) for cfn in constr_fns[i]]
            old_std = [st if adef not in defined else np.nan for st,adef in zip(std[i],alldef)]
            old_std += [np.nan]*(len(alldef)-len(old_std))
            old_chi = chi[i]
            
            # sort by original parameter name order
            old_std = [old_std[alldef.index(n)] for n in cgen.oldpar]
            
            # set to fitdata containers
            # [(parname),(par),(err),chi,fnpointer]
            data.set_fitresult([cgen.oldpar,old_par,old_std,old_chi,fnptrs[i]])
            
        # display in fit_files tab
        for key in fit_files.fit_lines:
            fit_files.fit_lines[key].show_fit_result()
        
        # show global chi
        fit_files.gchi_label['text'] = str(np.around(gchi,2))

        # do end-of-fit stuff
        fit_files.do_end_of_fit()
    
    # ====================================================================== #
    def do_parse(self,*args):
        """
            Detect new global variables
            returns split lines, new parameter names 
        """
        
        # clean input
        text = self.constr_text.split('\n')
        text = [t.strip() for t in text if '=' in t]
        
        # check for no input
        if not text:
            return
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # check for new parameters
        new_par = []
        for eq in eqn:
            lst = re.split('\W+',eq)    # split list non characters
            
            # throw out known things: numbers numpy equations
            delist = []
            for i,l in enumerate(lst):
                
                # check numpy functions
                if l == 'np':
                    delist.append(i)
                    delist.append(i+1)
                    continue
                
                # check integer
                try: 
                    int(l)
                except ValueError:
                    pass
                else:
                    delist.append(i)
                    continue
                
                # check variables
                if l in self.reserved_pars:  
                    delist.append(i)
                    continue
                    
            delist.sort()
            for i in delist[::-1]:
                try:
                    del lst[i]
                except IndexError:  # error raised on incomplete math: ex "np."
                    pass
                
            new_par.append(lst)

        # add result
        new_par = np.unique(np.concatenate(new_par))
        for k in new_par:
            
            # bad input
            if not k: continue
            
            # set defaults
            if k not in self.new_par['name'].values:
                self.new_par = self.new_par.append({'name':k,**self.default_parvals},
                                                   ignore_index=True)
        
        # drop results 
        for i,k in zip(self.new_par.index,self.new_par['name']):
            if k not in new_par:
                self.new_par.drop(i,inplace=True)
        
        # set fields
        self.new_par.sort_values('name',inplace=True)
        self.set_par_text()
        
        # logging
        self.logger.info('Parse found constraints for %s, and defined %s',
                         sorted(defined),
                         self.new_par['name'].values.tolist())
    
    # ====================================================================== #
    def get_input(self,*args):
        """Get input from text box."""
        self.constr_text = self.entry.get('1.0',END)
        self.do_parse()
        
    # ====================================================================== #
    def get_result_input(self,*args):
        """
            Set new_par row to match changes made by user
        """
    
        # get text
        try:
            text = {k:list(map(float,self.output_text[k].get('1.0',END).split('\n')[:-1])) \
                for k in self.output_text}
        # no update if blank
        except ValueError:
            return 
        
        # dataframe it
        try:
            text = pd.DataFrame(text)   
        # bad input
        except ValueError:
            return
        
        # get names of the parameters
        parnames = self.output_par_text.get('1.0',END).split('\n')[:-1]
        
        # update
        par = self.new_par.set_index('name')
        for i,name in enumerate(parnames):
            par.loc[name] = text.iloc[i]
        par.reset_index(inplace=True)
        self.new_par = par
        self.logger.debug('get_result_input: updated new_par')
    
    # ====================================================================== #
    def set_par_text(self):
        """
            Set the textboxes based on stored results in self.newpar
        """
        
        # get strings
        set_par = self.new_par.astype(str)
        
        # round
        numstr = '%'+('.%df' % self.bfit.rounding)
        for k in ('res','err'):
            set_par[k] = set_par.loc[:,k].apply(\
                    lambda x: numstr % np.around(float(x),self.bfit.rounding))
        
        # enable setting
        for k in ('res','err'):
            self.output_text[k].config(state='normal')
        self.output_par_text.config(state='normal')
        
        self.output_par_text.delete('1.0',END)
        self.output_par_text.insert(1.0,'\n'.join(set_par['name']))
        self.output_par_text_val = '\n'.join(set_par['name'])
        
        for k in self.output_text:
            self.output_text[k].delete('1.0',END)
            self.output_text[k].insert(1.0,'\n'.join(set_par[k]))
            self.output_text_val[k] = '\n'.join(set_par[k])
                
        # disable setting
        for k in ('res','err'):
            self.output_text[k].config(state='disabled')
        self.output_par_text.config(state='disabled')
        
    # ====================================================================== #
    def yview(self,*args):
        """
            Scrollbar for all output text fields
        """
        self.output_par_text.yview(*args)
        for k in self.output_text:
            self.output_text[k].yview(*args)



# Set up default fitting routines. 
# Derek Fujimoto
# Aug 2018

from bfit.fitting.fit_bdata import fit_list
import bfit.fitting.functions as fns
from bfit.fitting.decay_31mg import fa_31Mg
from functools import partial
import numpy as np
import bdata as bd

class fitter(object):

    # needed to tell users what routine this is
    __name__ = 'default'
    
    # Define possible fit functions for given run modes
    function_names = {  '20':('Exp','Bi Exp','Str Exp'),
                        '2h':('Exp','Bi Exp','Str Exp'),
                        '1f':('Lorentzian','Gaussian','BiLorentzian',),
                        '1w':('Lorentzian','Gaussian','BiLorentzian',),
                        '1n':('Lorentzian','Gaussian','BiLorentzian',),
                        '1e':('Lorentzian','Gaussian','BiLorentzian',),
                        '2e':('Lorentzian','Gaussian','BiLorentzian',)}
     
    # Define names of fit parameters:
    param_names = {     'Exp'       :('1_T1','amp'),
                        'Bi Exp'    :('1_T1','1_T1b','fraction_b','amp'),
                        'Str Exp'   :('1_T1','beta','amp'),
                        'Lorentzian':('peak','width','height','baseline'),
                        'BiLorentzian':('peak','widthA','heightA',
                                               'widthB','heightB','baseline'),
                        'Gaussian'  :('mean','sigma','height','baseline'),}

    # dictionary of initial parameters
    par_values = {}
    fn_list = {}
    epsilon = 1e-9  # for fixing parameters

    # define list of ok run modes 
    valid_asym_modes = ('c','p','n','sl_c','dif_c',)

    # ======================================================================= #
    def __init__(self,probe_species='8Li'):
        """
            probe_species: one of the keys in the bdata.life dictionary.
        """
        
        self.probe_species = probe_species
        
    # ======================================================================= #
    def __call__(self,fn_name,ncomp,data_list,hist_select,asym_mode,xlims):
        """
            Fitting controller. 
            
            fn_name: name of function to fit
            ncomp : number of components to incude (2 = biexp, for example)
            data_list: list of [[bdata object,pdict,doptions],]
            
                where pdict = {par:(init val,   # initial guess
                                    bound_lo,   # lower fitting bound
                                    bound_hi,   # upper fitting bound
                                    is_fixed,   # boolean, fix value?
                                    is_shared,  # boolean, share value globally?
                                   )
                              }
                where doptions = {  'omit':str,     # bins to omit in 1F calcs
                                    'rebin':int,    # rebinning factor
                                    'group':int,    # fitting group
                                 }
                                 
            hist_select: string for selection of histograms
            asym_mode:  input for asymmetry calculation type 
                            c: combined helicity
                            h: split helicity
                            p: positive helicity
                            n: negative helicity
                            
                        For 2e mode, prefix with:
                            sl_: combined timebins using slopes
                            dif_: combined timebins using differences
                            raw_: raw time-resolved
                            
                            ex: sl_c or raw_h or dif_c
            xlims: fit subrange of x axis
            returns dictionary of {run: [[par_names],[par_values],[par_errors],
                                        [chisquared],[fitfunction pointers]]}
                                   and global chisquared
        """

        # check ncomponents
        if ncomp < 1:
            raise RuntimeError('ncomp needs to be >= 1')
            
        asym_mode = asym_mode.replace('2e_','')
            
        # parameter names
        keylist = self.gen_param_names(fn_name,ncomp)
        npar = len(keylist)
        
        # gather list of data to fit 
        fn = []
        runs = []
        years = []
        p0 = []
        bounds = []
        omit = []
        rebin = []
        sharelist = np.zeros(npar,dtype=bool)
        fixedlist = []
        
        for data in data_list:
            
            # split data list into parts
            dat = data[0]
            pdict = data[1]
            doptions = data[2]
            
            # probe lifetime
            life = bd.life[self.probe_species]
            
            # get fitting function for 20 and 2h
            if dat.mode in ['20','2h']: 
                pulse = dat.get_pulse_s()
                
                # fit function
                fn1 = self.get_fn(fn_name,ncomp,pulse,life)
                
                # add corrections for probe daughters
                if self.probe_species == 'Mg31':
                    fn.append(lambda x,*par : fa_31Mg(x,pulse)*fn1(x,*par))
                else:
                    fn.append(fn1)
                
            # 1f functions
            else:                       
                fn.append(self.get_fn(fn_name,ncomp,-1,life))
            
            # get year and run
            runs.append(dat.run)
            years.append(dat.year)
            
            # get initial parameters
            p0.append(tuple(pdict[k][0] for k in keylist))
            
            # get fitting bounds
            bound = [[],[]]
            shlist = []
            fixed = []
            for k in keylist:
                
                # bounds
                bound[0].append(pdict[k][1])
                bound[1].append(pdict[k][2])
                
                # fixed
                fixed.append(pdict[k][3])    
                    
                # sharelist
                shlist.append(pdict[k][4])
                
            bounds.append(bound)
            fixedlist.append(fixed)
            
            # look for any sharelist
            sharelist += np.array(shlist)
            
            # rebin and omit
            try:
                rebin.append(doptions['rebin'])
            except KeyError:
                rebin.append(1)
                
            try:
                omit.append(doptions['omit'])
            except KeyError:
                omit.append('')                
            
        # fit data
        pars,covs,chis,gchi = fit_list(runs,years,fn,omit,rebin,sharelist,
                                       npar=npar,hist_select=hist_select,p0=p0,
                                       bounds=bounds,asym_mode=asym_mode,
                                       fixed=fixedlist,xlims=xlims)
        stds = [np.sqrt(np.diag(c)) for c in covs]
        
        # collect results
        return ({'.'.join(map(str,(d[0].year,d[0].run))):[keylist,p,s,c,f] \
                    for d,p,s,c,f in zip(data_list,pars,stds,chis,fn)},gchi)

    # ======================================================================= #
    def gen_param_names(self,fn_name,ncomp):
        """
            Make a list of the parameter names based on the number of components.
            
            fn_name: name of function (should match those in param_names)
            ncomp: number of components
            
            return (names)
        """
        
        # get names
        names_orig = self.param_names[fn_name]
        
        # special case of one component
        if ncomp == 1: 
            return names_orig
        
        # multicomponent: make copies of everything other than the baselines
        names = []
        for c in range(ncomp): 
            for n in names_orig:
                if 'base' in n: continue
                names.append(n+'_%d' % c)
                
        if 'base' in names_orig[-1]:
            names.append(names_orig[-1])
        
        return tuple(names)
        
    # ======================================================================= #
    def gen_init_par(self,fn_name,ncomp,bdataobj,asym_mode='combined'):
        """Generate initial parameters for a given function.
        
            fname: name of function. Should be the same as the param_names keys
            ncomp: number of components
            bdataobj: a bdata object representative of the fitting group. 
            asym_mode: what kind of asymmetry to fit
            
            Set and return dictionary of initial parameters. 
                {par_name:par_value}
        """
        
        # asym_mode un-used types 
        if asym_mode in (   'h',           # Split Helicity          
                            'hm',          # Matched Helicity
                            'hs',          # Shifted Split          
                            'cs',          # Shifted Combined      
                            'hp',          # Matched Peak Finding
                            'r',           # Raw Scans            
                            'rhist',       # Raw Histograms          
                            '2e_raw_c',    # Combined Hel Raw       
                            '2e_raw_h',    # Split Hel Raw          
                            '2e_sl_h',     # Split Hel Slopes
                            '2e_dif_h',    # Split Hel Diff           
                            'ad',          # Alpha Diffusion  
                            "at_c",        # Combined Hel (Alpha Tag)
                            "at_h",        # Split Hel (Alpha Tag) 
                            "nat_c",       # Combined Hel (!Alpha Tag)
                            "nat_h",       # Split Hel (!Alpha Tag)
                        ):
            errmsg = "Asymmetry calculation type not implemented for fitting"
            raise RuntimeError(errmsg)
        
        # get asymmetry
        x,a,da = bdataobj.asym(asym_mode)
        
        # set pulsed exp fit initial parameters
        if fn_name in ('Exp','Bi Exp','Str Exp'):
            # ampltitude average of first 5 bins
            amp = abs(np.mean(a[0:5])/ncomp)
            
            # T1: time after beam off to reach 1/e
            idx = int(bdataobj.ppg.beam_on.mean)
            beam_duration = x[idx]
            amp_beamoff = a[idx]
            target = amp_beamoff/np.exp(1)
            
            x_target = x[np.sum(a>target)]
            T1 = abs(x_target-beam_duration)
            
            # bounds and amp
            if asym_mode == 'n':
                amp_bounds = (-np.inf,np.inf)
                amp = -amp
            else:
                amp_bounds = (0,np.inf)
            
            # set values
            par_values = {'amp':(amp,*amp_bounds),
                          '1_T1':(1./T1,0,np.inf),
                          '1_T1b':(1./T1,0,np.inf),
                          'fraction_b':(0.5,0,1),
                          'beta':(0.5,0,1)}
                         
        # set time integrated fit initial parameters
        elif fn_name in ('Lorentzian','Gaussian','BiLorentzian'):
            
            # get peak asym value
            amin = min(a[a!=0])
            
            peak = x[np.where(a==amin)[0][0]]
            base = np.mean(a[:5])
            height = base-amin
            width = 2*abs(peak-x[np.where(a<amin+height/2)[0][0]])
            
            # bounds
            if asym_mode == 'n':
                height_bounds = (-np.inf,0)
            else:
                height_bounds = (0,np.inf)
            
            # set values
            if fn_name == 'Lorentzian':
                par_values = {'peak':(peak,min(x),max(x)),
                              'width':(width,0,np.inf),
                              'height':(height,*height_bounds),
                              'baseline':(base,-np.inf,np.inf)
                             }
            elif fn_name == 'Gaussian':
                par_values = {'mean':(peak,min(x),max(x)),
                              'sigma':(width,0,np.inf),
                              'height':(height,*height_bounds),
                              'baseline':(base,-np.inf,np.inf)
                              }
            if fn_name == 'BiLorentzian':
                par_values = {'peak':(peak,min(x),max(x)),
                              'widthA':(width,0,np.inf),
                              'heightA':(height,*height_bounds),
                              'widthB':(width,0,np.inf),
                              'heightB':(height,*height_bounds),
                              'baseline':(base,-np.inf,np.inf)
                             }
        else:
            raise RuntimeError('Bad function name.')
        
        # do multicomponent
        par_values2 = {}
        if ncomp > 1: 
            for c in range(ncomp): 
                for n in par_values.keys():
                    if 'baseline' not in n:
                        par_values2[n+'_%d' % c] = par_values[n]
                    else:
                        par_values2[n] = par_values[n]
        else:
            par_values2 = par_values
            
        return par_values2
        
    # ======================================================================= #
    def get_fn(self,fn_name,ncomp=1,pulse_len=-1,lifetime=-1):
        """
            Get the fitting function used.
            
                fn_name: string of the function name users will select. 
                ncomp: number of components, ex if 2, then return exp+exp
                pulse_len: duration of beam on in s
                lifetime: lifetime of probe in s
            
            Returns python function(x,*pars)
        """
        
        # set fitting function
        if fn_name == 'Lorentzian':
            fn =  fns.lorentzian
            self.mode=1
        elif fn_name == 'BiLorentzian':
            fn =  fns.bilorentzian
            self.mode=1
        elif fn_name == 'Gaussian':
            fn =  fns.gaussian
            self.mode=1
        elif fn_name == 'Exp':
            fn =  fns.pulsed_exp(lifetime,pulse_len)
            self.mode=2
        elif fn_name == 'Bi Exp':
            fn =  fns.pulsed_biexp(lifetime,pulse_len)
            self.mode=2
        elif fn_name == 'Str Exp':
            fn =  fns.pulsed_strexp(lifetime,pulse_len)
            self.mode=2
        else:
            raise RuntimeError('Fitting function not found.')
        
        # Make final function based on number of components
        fnlist = [fn]*ncomp
        
        if self.mode == 1:
            fnlist.append(lambda x,b: b)
        fn = fns.get_fn_superpos(fnlist)
        
        return fn
        
        
        
        
        



# Track which plots are active, which to draw in. 
# Derek Fujimoto
# June 2019

import matplotlib.pyplot as plt

# =========================================================================== #
class PltTracker(object):
    """
        active:         dictionary, id number of active plot
        plots:          dictionary, list of plots drawn for type
    """
    
    # ======================================================================= #
    def __init__(self):
        
        # lists for tracking all plots 
        self.plots = {'data':[],'fit':[],'param':[],'periodic':[]}
        
        # track the active plot 
        self.active = {'data':0,'fit':0,'param':0,'periodic':0}
    
    # ======================================================================= #
    def _close_figure(self,event):
        """Remove figure from list"""
        
        # get number and style
        number = event.canvas.figure.number
        style = event.canvas.style
        
        # disconnect events
        event.canvas.mpl_disconnect(event.canvas.user_close)
        event.canvas.mpl_disconnect(event.canvas.user_active)
        
        # close the winow
        plt.figure(number).clf()
        plt.close(number)
        
        # remove from list 
        self.plots[style].remove(number)
        
        # reset active
        try:
            self.active[style] = self.plots[style][-1]
        except IndexError:
            self.active[style] = 0
                        
    # ======================================================================= #
    def _decorator(self,style,fn,*args,**kwargs):
        """
            Function wrapper
            
            style: one of "data", "fit", "param"
            fn: matplotlib function to operate on 
            args: passed to fn
            kwargs: passed to fn
        """
        
        # switch 
        plt.figure(self.active[style])

        # run function 
        output = fn(*args,**kwargs)
        
        return output
    
    # ======================================================================= #
    def _update_active_id(self,event):
        """
            Update the active figure id based on click event.
        """
        number = event.canvas.figure.number
        style = event.canvas.style
        self.active[style] = number
    
    # ======================================================================= #
    def annotate(self,style,*args,**kwargs):
        return self._decorator(style,plt.annotate,*args,**kwargs)
    
    # ======================================================================= #
    def axhline(self,style,*args,**kwargs):
        return self._decorator(style,plt.axhline,*args,**kwargs)
    
    # ======================================================================= #
    def axvline(self,style,*args,**kwargs):
        return self._decorator(style,plt.axvline,*args,**kwargs)
            
    # ======================================================================= #
    def clf(self,style):
        """Get the current axes for the style"""
        return self._decorator(style,plt.clf)
        
    # ======================================================================= # 
    def errorbar(self, style, x, y, yerr=None, xerr=None, fmt='', ecolor=None, 
                 elinewidth=None, capsize=None, barsabove=False, lolims=False, 
                 uplims=False, xlolims=False, xuplims=False, errorevery=1, 
                 capthick=None, *, data=None, **kwargs):
        """
            Plot data.
            
            style: one of "data", "fit", or "param"
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active[style]
        
        # make new figure if needed 
        if active_style == 0:   
            self.figure(style)
            active_style = self.active[style]
        
        # draw in active style 
        plt.figure(active_style)
        obj = plt.errorbar(x, y, yerr=yerr, xerr=xerr, fmt=fmt, ecolor=ecolor, 
                     elinewidth=elinewidth, capsize=capsize, 
                     barsabove=barsabove, lolims=lolims, uplims=uplims, 
                     xlolims=xlolims, xuplims=xuplims, errorevery=errorevery, 
                     capthick=capthick, data=data, **kwargs)
        
        return obj
    
    # ======================================================================= #
    def figure(self,style,**kwargs):
        """
            Make new figure.
            
            style: one of "data", "fit", or "param"
            kwargs: keyword arguments to pass to plt.figure
        """
        
        # make figure
        fig = plt.figure(**kwargs)
        
        # make events and save as canvas attribute
        fig.canvas.user_close = fig.canvas.mpl_connect('close_event', self._close_figure)
        fig.canvas.user_active = fig.canvas.mpl_connect('button_press_event', self._update_active_id)
        
        # set style
        fig.canvas.style = style
        
        # set window name 
        fig.canvas.set_window_title('Figure %d (%s)' % (fig.number,style.title()))
        
        # update lists
        self.plots[style].append(fig.number)
        self.active[style] = fig.number
        
        return fig

    # ======================================================================= #
    def gca(self,style):
        if not self.plots[style]: self.figure(style)
        return self._decorator(style,plt.gca)
    
    # ======================================================================= #
    def gcf(self,style):
        if not self.plots[style]: self.figure(style)
        return self._decorator(style,plt.gcf)
    
    # ======================================================================= #
    def legend(self,style,*args,**kwargs):
        self._decorator(style,plt.legend,*args,**kwargs)
        
    # ======================================================================= #
    def plot(self,style, *args, scalex=True, scaley=True, data=None, **kwargs):
        """
            Plot data.
            
            style: one of "data", "fit", or "param"
            other arguments: defaults for matplotlib.pyplot.plot
        """
        
        # get active for this style
        active_style = self.active[style]
        
        # make new figure if needed 
        if active_style not in self.plots[style]:   
            self.figure(style)
        
        # draw in active style 
        plt.figure(active_style)
        obj = plt.plot(*args,scalex=scalex, scaley=scaley, data=data,**kwargs)
        
        return obj

    # ======================================================================= #
    def text(self,style,*args,**kwargs):
        return self._decorator(style,plt.text,*args,**kwargs)

    # ======================================================================= #
    def tight_layout(self,style,*args,**kwargs):
        return self._decorator(style,plt.tight_layout,*args,**kwargs)

    # ======================================================================= #
    def xlabel(self,style,*args,**kwargs):
        return self._decorator(style,plt.xlabel,*args,**kwargs)
    
    # ======================================================================= #
    def xlim(self,style,*args,**kwargs):
        return self._decorator(style,plt.xlim,*args,**kwargs)
    
    # ======================================================================= #
    def xticks(self,style,*args,**kwargs):
        return self._decorator(style,plt.xticks,*args,**kwargs)
    
    # ======================================================================= #
    def ylabel(self,style,*args,**kwargs):
        return self._decorator(style,plt.ylabel,*args,**kwargs)
    
    # ======================================================================= #
    def ylim(self,style,*args,**kwargs):
        return self._decorator(style,plt.ylim,*args,**kwargs)
    
    # ======================================================================= #
    def yticks(self,style,*args,**kwargs):
        return self._decorator(style,plt.yticks,*args,**kwargs)
    
    # ======================================================================= #
    def zlabel(self,style,*args,**kwargs):
        return self._decorator(style,plt.zlabel,*args,**kwargs)
    
    # ======================================================================= #
    def zlim(self,style,*args,**kwargs):
        return self._decorator(style,plt.zlim,*args,**kwargs)
    
    # ======================================================================= #
    def zticks(self,style,*args,**kwargs):
        return self._decorator(style,plt.zticks,*args,**kwargs)
    



# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox
from functools import partial
import logging
import numpy as np
from scipy.optimize import curve_fit

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout,on_entry_click
from bfit.backend.get_model import get_model
import bfit.backend.colors as colors
from bfit.backend.raise_window import raise_window

# ========================================================================== #
class popup_fit_results(object):
    """
        Popup window for modelling the fit results with a function
        
        bfit
        fittab
        logger
        
        model_chi_label:    Label, chisquared output
        model_fn:           Funtion handle, fit model for pars, created on fit
        model_p0:           Initial parameters for model fitting
        model_entry:        Entry: model function
        model_results:      Tuple, (par,std) for model fit results
        model_results_fn:   StringVar, fit function to model results
        model_results_par:  StringVar, parameter list of model_results fn
        model_results_text: Text, output model results
        model_errors_text:  Text, output model result errors
        modelp0_entry:      Entry: model p0 list
        modelpar_entry:     Entry: model parameter list
            
        win:                Toplevel
        
    """

    # ====================================================================== #
    def __init__(self,bfit,par_str='a,b',fn_str='a*x+b',p0_str='1,1',
                 result_str='',error_str=''):
        self.bfit = bfit
        self.fittab = bfit.fit_files
        
        self.par_str = par_str
        self.fn_str = fn_str
        self.p0_str = p0_str
        self.result_str = result_str
        self.error_str = error_str
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title('Model the Fit Results')
        fit_fitresults_frame = ttk.Frame(self.win,relief='sunken',pad=5)
        
        # icon
        self.bfit.set_icon(self.win)

        # Key bindings
        self.win.bind('<Return>',self.do_fit_model)             
        self.win.bind('<KP_Enter>',self.do_fit_model)
        
        self.model_results_fn = StringVar()
        self.model_results_par = StringVar()
        self.model_p0 = StringVar()
        
        # entry
        model_entry_frame = Frame(fit_fitresults_frame)
        self.modelpar_entry = ttk.Entry(model_entry_frame,
                                  textvariable=self.model_results_par,width=6)
        
        self.model_entry = ttk.Entry(model_entry_frame,
                                     textvariable=self.model_results_fn,width=26)
        
        self.modelp0_entry = ttk.Entry(model_entry_frame,
                                  textvariable=self.model_p0,width=26)
                                     
        # entry defaults (parameters)
        self.modelpar_entry.insert(0,self.par_str)
        entry_parfn = partial(on_entry_click,text=self.par_str,entry=self.modelpar_entry)
        on_focusout_parfn = partial(on_focusout,text=self.par_str,entry=self.modelpar_entry)
        self.modelpar_entry.bind('<FocusIn>', entry_parfn)
        self.modelpar_entry.bind('<FocusOut>', on_focusout_parfn)
        self.modelpar_entry.config(foreground=colors.entry_grey)
       
        # entry defaults (function)
        self.model_entry.insert(0,self.fn_str)
        entry_fnfn = partial(on_entry_click,text=self.fn_str,entry=self.model_entry)
        on_focusout_fnfn = partial(on_focusout,text=self.fn_str,entry=self.model_entry)
        self.model_entry.bind('<FocusIn>', entry_fnfn)
        self.model_entry.bind('<FocusOut>', on_focusout_fnfn)
        self.model_entry.config(foreground=colors.entry_grey)
       
        # entry defaults (p0)
        self.modelp0_entry.insert(0,self.p0_str)
        entry_fnp0 = partial(on_entry_click,text=self.p0_str,entry=self.modelp0_entry)
        on_focusout_fnp0 = partial(on_focusout,text=self.p0_str,entry=self.modelp0_entry)
        self.modelp0_entry.bind('<FocusIn>', entry_fnp0)
        self.modelp0_entry.bind('<FocusOut>', on_focusout_fnp0)
        self.modelp0_entry.config(foreground=colors.entry_grey)
       
        # buttons
        model_fit_button = ttk.Button(fit_fitresults_frame,text='Fit',command=self.do_fit_model)
        
        # chisq
        self.model_chi_label = ttk.Label(fit_fitresults_frame,text='',justify=LEFT)
        
        # text for output
        self.model_results_text = Text(fit_fitresults_frame,width=17,height=8,state='normal')
        self.model_errors_text = Text(fit_fitresults_frame,width=17,height=8,state='normal')
        
        self.model_results_text.insert('1.0',self.result_str)
        self.model_errors_text.insert('1.0',self.error_str)
        
        model_entry_frame.grid(column=0,row=0,columnspan=2,sticky=W,pady=2)
        
        ttk.Label(model_entry_frame,text="Param").grid(column=0,row=0,sticky=(N,W))
        ttk.Label(model_entry_frame,text="Model").grid(column=2,row=0,sticky=(N,W))
        self.modelpar_entry.grid(column=0,row=1,padx=2,pady=1,sticky=(N,W))
        ttk.Label(model_entry_frame,text=":").grid(column=1,row=1,sticky=(N,W))
        self.model_entry.grid(column=2,row=1,padx=2,pady=1,sticky=(N,W))
        ttk.Label(model_entry_frame,text="P0").grid(column=0,row=2,sticky=(E))
        ttk.Label(model_entry_frame,text=":").grid(column=1,row=2,sticky=(N,W))
        self.modelp0_entry.grid(column=2,row=2,padx=2,pady=1,sticky=(N,W))
        
        model_fit_button.grid(column=0,row=2,padx=2,pady=2)
        self.model_chi_label.grid(column=1,row=2,padx=2,pady=2)
        
        ttk.Label(fit_fitresults_frame,text="Results").grid(column=0,row=3,pady=2,sticky=N)
        ttk.Label(fit_fitresults_frame,text="Errors").grid(column=1,row=3,pady=2,sticky=N)
        
        self.model_results_text.grid(column=0,row=4,padx=2)
        self.model_errors_text.grid(column=1,row=4,padx=2)
                
        # grid frame
        fit_fitresults_frame.grid(column=0,row=0)
        self.logger.debug('Initialization success. Starting mainloop.')
    
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()
    
    # ====================================================================== #
    def do_fit_model(self,*args):
        
        # get fit data
        xstr = self.fittab.xaxis.get()
        ystr = self.fittab.yaxis.get()
        
        # Make model 
        self.par_str = self.model_results_par.get()
        parstr = self.par_str
        parlst = parstr.split(',')
        npar = len(parlst)
        
        if parstr[-1] == ',': parstr = parstr[:-1]
        self.fn_str = self.model_results_fn.get()
        model = 'lambda x,%s : %s' % (parstr,self.fn_str)
        
        self.logger.info('Fitting model %s for x="%s", y="%s"',model,xstr,ystr)
        
        model = get_model(model) 
        self.model_fn = model
        npar = len(parstr.split(','))
        
        # get data values
        try:
            xvals, xerrs = self.fittab.get_values(xstr)
            yvals, yerrs = self.fittab.get_values(ystr)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter selection')
            messagebox.showerror("Error",'Select two input parameters')
            raise err
        except (KeyError,AttributeError) as err:
            self.logger.error('Parameter "%s or "%s" not found for fitting',
                              xstr,ystr)
            messagebox.showerror("Error",
                    'Parameter "%s" or "%s" not found' % (xstr,ystr))
            raise err
            
        xvals = np.asarray(xvals)
        yvals = np.asarray(yvals)
        yerrs = np.asarray(yerrs)
            
        # get p0
        self.p0_str = self.model_p0.get()
        p0 = self.p0_str
        if p0[-1] == ',': p0 = p0[:-1]
        
        try:
            if p0:  p0 = list(map(float,p0.split(',')))
            else:   p0 = np.ones(npar)
        except ValueError:
            msg = 'Bad p0 input: use format p0,p1,p2,...'
            messagebox.showerror('Error',msg)
            self.logging.error('Bad p0 input')
            raise ValueError(msg) from None
        
        if len(p0) < npar:  p0 = np.concatenate((p0,np.ones(npar-len(p0))))
        if len(p0) > npar:  p0 = p0[:npar]
            
        # fit model 
        if all(np.isnan(yerrs)): yerrs = None
        
        par,cov = curve_fit(model,xvals,yvals,sigma=yerrs,absolute_sigma=True,p0=p0)
        std = np.diag(cov)**0.5
        
        if yerrs is None:   yerrs = np.ones(len(xvals))
        chi = np.sum(((model(xvals,*par)-yvals)/yerrs)**2)/(len(xvals)-npar)
        
        # display results 
        number = '%'+('%.df' % self.bfit.rounding)
        outtext = [p+': '+number % r for p,r in zip(parlst,par)]
        self.model_results_text.delete('1.0',END)
        self.result_str = '\n'.join(outtext)
        self.model_results_text.insert('1.0',self.result_str)
        
        outtext = [number % r for r in std]
        self.model_errors_text.delete('1.0',END)
        self.error_str = '\n'.join(outtext)
        self.model_errors_text.insert('1.0',self.error_str)
        
        self.model_chi_label['text'] = 'ChiSq: %.2f' % np.around(chi,2)
        
        self.logger.info('Fit model results: %s, Errors: %s',str(par),str(std))
        
        self.model_results = (par,std)
        self.fittab.draw_param()
        self.draw_model()    
        
    # ======================================================================= #
    def draw_model(self,*args):
        figstyle = 'param'
        
        # get draw components
        xstr = self.fittab.xaxis.get()
        ystr = self.fittab.yaxis.get()
        
        self.logger.info('Draw model parameters "%s" vs "%s"',ystr,xstr)
        
        # get data values
        try:
            xvals, xerrs = self.fittab.get_values(xstr)
            yvals, yerrs = self.fittab.get_values(ystr)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter selection')
            messagebox.showerror("Error",'Select two input parameters')
            raise err from None
        except (KeyError,AttributeError) as err:
            self.logger.error('Parameter "%s or "%s" not found for drawing model',
                              xstr,ystr)
            messagebox.showerror("Error",
                    'Parameter "%s" or "%s" not found' % (xstr,ystr))
            raise err from None

        # get fit function
        fn = self.model_fn

        # get draw style
        style = self.bfit.draw_style.get()
        
        if style == 'new':
            self.fittab.plt.figure(figstyle)
        elif style == 'redraw':
            self.fittab.plt.clf(figstyle)
        
        self.fittab.plt.gca(figstyle)
            
        # draw
        fitx = np.linspace(min(xvals),max(xvals),self.fittab.n_fitx_pts)
        f = self.fittab.plt.plot(figstyle,fitx,fn(fitx,*self.model_results[0]),color='k')
        
        # plot elements
        self.fittab.plt.xlabel(figstyle,xstr)
        self.fittab.plt.ylabel(figstyle,ystr)
        self.fittab.plt.tight_layout(figstyle)
        
        raise_window()
    



# fetch_files tab for bfit
# Derek Fujimoto
# Nov 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
from bfit import logger_name
from bdata import bdata
from functools import partial
from bfit.backend.fitdata import fitdata
from bfit.backend.entry_color_set import on_focusout,on_entry_click
import bfit.backend.colors as colors
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import time,datetime,os,logging,glob

__doc__="""
    """

# =========================================================================== #
# =========================================================================== #
class fetch_files(object):
    """
        Data fields:
            
            asym_type: drawing style
            canvas_frame_id: id number of frame in canvas
            check_rebin: IntVar for handling rebin aspect of checkall
            check_bin_remove: StringVar for handing omission of 1F data
            check_state: BooleanVar for handling check all
            check_state_data: BooleanVar for handling check_all_data
            check_state_fit: BooleanVar for handling check_all_fit
            check_state_res: BooleanVar for handling check_all_res
            
            data_canvas: canvas object allowing for scrolling 
            dataline_frame: frame holding all the data lines. Exists as a window
                in the data_canvas
            
            entry_asym_type: combobox for asym calc and draw type
            year: IntVar of year to fetch runs from 
            run: StringVar input to fetch runs.
            bfit: pointer to parent class
            data_lines: dictionary of dataline obj, keyed by run number
            data_lines_old: dictionary of removed dataline obj, keyed by run number
            fet_entry_frame: frame of fetch tab
            runmode_label: display run mode
            runmode: display run mode string
            max_number_fetched: max number of files you can fetch
    """
    
    runmode_relabel = {'20':'SLR (20)',
                       '1f':'Frequency Scan (1f)',
                       '1w':'Frequency Comb (1w)',
                       '2e':'Random Freq. (2e)',
                       '1n':'Rb Cell Scan (1n)',
                       '1e':'Field Scan (1e)',
                       '2h':'Alpha Tagged (2h)'}
    run_number_starter_line = '40001 40005-40010 (run numbers)'
    bin_remove_starter_line = '24 100-200 (bins)'
    max_number_fetched = 500
    
    # ======================================================================= #
    def __init__(self,fetch_data_tab,bfit):
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing')
    
        # initialize
        self.bfit = bfit
        self.data_lines = {}
        self.data_lines_old = {}
        self.fit_input_tabs = {}
        self.check_rebin = IntVar()
        self.check_bin_remove = StringVar()
        self.check_state = BooleanVar()
        self.fetch_data_tab = fetch_data_tab
        
        # Frame for specifying files -----------------------------------------
        fet_entry_frame = ttk.Labelframe(fetch_data_tab,text='Specify Files')
        self.year = IntVar()
        self.run = StringVar()
        
        self.year.set(self.bfit.get_latest_year())
        
        entry_year = Spinbox(fet_entry_frame,textvariable=self.year,width=5,
                             from_=2000,to=datetime.datetime.today().year)
        entry_run = ttk.Entry(fet_entry_frame,textvariable=self.run,width=85)
        entry_run.insert(0,self.run_number_starter_line)
        entry_fn = partial(on_entry_click,text=self.run_number_starter_line,\
                            entry=entry_run)
        on_focusout_fn = partial(on_focusout,text=self.run_number_starter_line,\
                            entry=entry_run)
        entry_run.bind('<FocusIn>', entry_fn)
        entry_run.bind('<FocusOut>', on_focusout_fn)
        entry_run.config(foreground=colors.entry_grey)
        
        # fetch button
        fetch = ttk.Button(fet_entry_frame,text='Fetch',command=self.get_data)
        
        # grid and labels
        fet_entry_frame.grid(column=0,row=0,sticky=(N,W,E),columnspan=2,padx=5,pady=5)
        ttk.Label(fet_entry_frame,text="Year:").grid(column=0,row=0,sticky=W)
        entry_year.grid(column=1,row=0,sticky=(W))
        ttk.Label(fet_entry_frame,text="Run Number:").grid(column=2,row=0,sticky=W)
        entry_run.grid(column=3,row=0,sticky=W)
        fetch.grid(column=4,row=0,sticky=E)
        
        # padding 
        for child in fet_entry_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)
        
        # Frame for run mode -------------------------------------------------
        runmode_label_frame = ttk.Labelframe(fetch_data_tab,pad=(10,5,10,5),\
                text='Run Mode',)
        
        self.runmode_label = ttk.Label(runmode_label_frame,text="",justify=CENTER)
        
        # Scrolling frame to hold datalines
        yscrollbar = ttk.Scrollbar(fetch_data_tab, orient=VERTICAL)         
        self.data_canvas = Canvas(fetch_data_tab,bd=0,              # make a canvas for scrolling
                yscrollcommand=yscrollbar.set,                      # scroll command receive
                scrollregion=(0, 0, 5000, 5000),confine=True)       # default size
        yscrollbar.config(command=self.data_canvas.yview)           # scroll command send
        dataline_frame = ttk.Frame(self.data_canvas,pad=5)          # holds 
        
        self.canvas_frame_id = self.data_canvas.create_window((0,0),    # make window which can scroll
                window=dataline_frame,
                anchor='nw')
        dataline_frame.bind("<Configure>",self.config_canvas) # bind resize to alter scrollable region
        self.data_canvas.bind("<Configure>",self.config_dataline_frame) # bind resize to change size of contained frame
        
        # Frame to hold everything on the right ------------------------------
        bigright_frame = ttk.Frame(fetch_data_tab,pad=5)
        
        # Frame for group set options ----------------------------------------
        right_frame = ttk.Labelframe(bigright_frame,\
                text='Operations on Checked Items',pad=30)
        
        check_remove = ttk.Button(right_frame,text='Remove',\
                command=self.remove_all,pad=5)
        check_draw = ttk.Button(right_frame,text='Draw',\
                command=lambda:self.draw_all('data'),pad=5)
        
        check_rebin_label = ttk.Label(right_frame,text="SLR Rebin:",pad=5)
        check_rebin_box = Spinbox(right_frame,from_=1,to=100,width=3,\
                textvariable=self.check_rebin,
                command=self.set_all)
        check_bin_remove_entry = ttk.Entry(right_frame,\
                textvariable=self.check_bin_remove,width=20)
        
        check_all_box = ttk.Checkbutton(right_frame,
                text='Force Check State',variable=self.check_state,
                onvalue=True,offvalue=False,pad=5,command=self.check_all)
        self.check_state.set(True)
        
        right_checkbox_frame = ttk.Frame(right_frame)
        
        self.check_state_data = BooleanVar()        
        check_data_box = ttk.Checkbutton(right_checkbox_frame,
                text='Data',variable=self.check_state_data,
                onvalue=True,offvalue=False,pad=5,command=self.check_all_data)
        self.check_state_data.set(True)
        
        self.check_state_fit = BooleanVar()        
        check_fit_box = ttk.Checkbutton(right_checkbox_frame,
                text='Fit',variable=self.check_state_fit,
                onvalue=True,offvalue=False,pad=5,command=self.check_all_fit)
        
        self.check_state_res = BooleanVar()        
        check_res_box = ttk.Checkbutton(right_checkbox_frame,
                text='Res',variable=self.check_state_res,
                onvalue=True,offvalue=False,pad=5,command=self.check_all_res)
                
        check_toggle_button = ttk.Button(right_frame,\
                text='Toggle All Check States',command=self.toggle_all,pad=5)
        
        # add grey to check_bin_remove_entry
        check_bin_remove_entry.insert(0,self.bin_remove_starter_line)
        
        check_entry_fn = partial(on_entry_click,\
                text=self.bin_remove_starter_line,\
                entry=check_bin_remove_entry)
        
        check_on_focusout_fn = partial(on_focusout,\
                text=self.bin_remove_starter_line,\
                entry=check_bin_remove_entry)
        
        check_bin_remove_entry.bind('<FocusIn>', check_entry_fn)
        check_bin_remove_entry.bind('<FocusOut>', check_on_focusout_fn)
        check_bin_remove_entry.config(foreground=colors.entry_grey)
                
        # grid
        runmode_label_frame.grid(column=2,row=0,sticky=(N,W,E,S),pady=5,padx=5)
        self.runmode_label.grid(column=0,row=0,sticky=(N,W,E))
        
        bigright_frame.grid(column=2,row=1,sticky=(N,E))
        
        self.data_canvas.grid(column=0,row=1,sticky=(E,W,S,N),padx=5,pady=5)
        yscrollbar.grid(column=1,row=1,sticky=(W,S,N),pady=5)
        
        check_data_box.grid(        column=0,row=0,sticky=(N))
        check_fit_box.grid(         column=1,row=0,sticky=(N))
        check_res_box.grid(         column=2,row=0,sticky=(N)) 
        
        right_frame.grid(           column=0,row=0,sticky=(N,E,W))
        r = 0
        check_all_box.grid(         column=0,row=r,sticky=(N),columnspan=2); r+= 1
        right_checkbox_frame.grid(  column=0,row=r,sticky=(N),columnspan=2); r+= 1
        check_toggle_button.grid(   column=0,row=r,sticky=(N,E,W),columnspan=2,pady=1,padx=5); r+= 1
        check_draw.grid(            column=0,row=r,sticky=(N,W,E),pady=5,padx=5);
        check_remove.grid(          column=1,row=r,sticky=(N,E,W),pady=5,padx=5); r+= 1
        check_rebin_label.grid(     column=0,row=r)
        check_rebin_box.grid(       column=1,row=r); r+= 1
        check_bin_remove_entry.grid(column=0,row=r,sticky=(N),columnspan=2); r+= 1
        bigright_frame.grid(        rowspan=2,sticky=(N,E,W))
        
        check_rebin_box.grid_configure(padx=5,pady=5,sticky=(E,W))
        check_rebin_label.grid_configure(padx=5,pady=5,sticky=(E,W))
        
        # resizing
        fetch_data_tab.grid_columnconfigure(0, weight=1)        # main area
        fetch_data_tab.grid_rowconfigure(1,weight=1)            # main area
        
        for i in range(3):
            if i%2 == 0:    fet_entry_frame.grid_columnconfigure(i, weight=2)
        fet_entry_frame.grid_columnconfigure(3, weight=1)
            
        self.data_canvas.grid_columnconfigure(0,weight=1)    # fetch frame 
        self.data_canvas.grid_rowconfigure(0,weight=1)
            
        # asymmetry calculation
        style_frame = ttk.Labelframe(bigright_frame,text='Asymmetry Calculation',\
                pad=5)
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(style_frame,\
                textvariable=self.asym_type,state='readonly',\
                width=20)
        self.entry_asym_type['values'] = ()
        
        style_frame.grid(column=0,row=1,sticky=(W,N,E))
        style_frame.grid_columnconfigure(0,weight=1)
        self.entry_asym_type.grid(column=0,row=0,sticky=(N,E,W),padx=10)
        
        # passing
        self.entry_run = entry_run
        self.entry_year = entry_year
        self.check_rebin_box = check_rebin_box
        self.check_bin_remove_entry = check_bin_remove_entry
        self.check_all_box = check_all_box
        self.dataline_frame = dataline_frame

        self.logger.debug('Initialization success.')
    
    # ======================================================================= #
    def __del__(self):
        
        # delete lists and dictionaries
        if hasattr(self,'data_lines'):      del self.data_lines
        if hasattr(self,'data_lines_old'):  del self.data_lines_old
        
        # kill buttons and frame
        try:
            for child in self.fetch_data_tab.winfo_children():
                child.destroy()
            self.fetch_data_tab.destroy()
        except Exception:
            pass
        
    # ======================================================================= #
    def _do_check_all(self,state,var,box):
        """
            Force all tickboxes of a given type to be in a given state, assuming 
            the tickbox is active
        """
        
        self.logger.info('Changing state of all %s tickboxes to %s',var,state)
        for k in self.data_lines.keys():
            
            # check if tickbox is object variable
            if hasattr(self.data_lines[k],box):
                
                # check if tickbox is disabled
                if str(getattr(self.data_lines[k],box)['state']) == 'disabled':
                    continue
                    
            # set value
            getattr(self.data_lines[k],var).set(state)

    # ======================================================================= #
    def canvas_scroll(self,event):
        """Scroll canvas with files selected."""
        if event.num == 4:
            self.data_canvas.yview_scroll(-1,"units")
        elif event.num == 5:
            self.data_canvas.yview_scroll(1,"units")
    
    # ======================================================================= #
    def check_all(self):  
        self._do_check_all(self.check_state.get(),'check_state','None')
        
    def check_all_data(self):  
        self._do_check_all(self.check_state_data.get(),'check_data','None')
        
    def check_all_fit(self):  
        self._do_check_all(self.check_state_fit.get(),'check_fit','draw_fit_checkbox')
    
    def check_all_res(self):  
        self._do_check_all(self.check_state_res.get(),'check_res','draw_res_checkbox')    
        
    # ======================================================================= #
    def config_canvas(self,event):
        """Alter scrollable region based on canvas bounding box size. 
        (changes scrollbar properties)"""
        self.data_canvas.configure(scrollregion=self.data_canvas.bbox("all"))
    
    # ======================================================================= #
    def config_dataline_frame(self,event):
        """Alter size of contained frame in canvas. Allows for inside window to 
        be resized with mouse drag""" 
        self.data_canvas.itemconfig(self.canvas_frame_id,width=event.width)
        
    # ======================================================================= #
    def draw_all(self,figstyle,ignore_check=False):
        """
            Draw all data in data lines
            
            figstyle: one of "data", "fit", or "param" to choose which figure 
                    to draw in
            ignore_check: draw all with no regard to whether the run has been 
                    selected
        """
        
        self.logger.debug('Drawing all data (ignore check: %s)', ignore_check)
        
        # condense drawing into a funtion
        def draw_lines():
            for i,r in enumerate(self.data_lines.keys()):
                if self.data_lines[r].check_state.get() or ignore_check:
                    self.data_lines[r].draw(figstyle)
                
        # get draw style
        style = self.bfit.draw_style.get()
        self.logger.debug('Draw style: "%s"',style)
        
        # make new figure, draw stacked
        if style == 'stack':
            draw_lines()
            
        # overdraw in current figure, stacked
        elif style == 'redraw':
            self.bfit.draw_style.set('stack')
            
            if self.bfit.plt.plots[figstyle]:
                self.bfit.plt.clf(figstyle)
            
            draw_lines()
            self.bfit.draw_style.set('redraw')
            
        # make new figure, draw single
        elif style == 'new':
            self.bfit.draw_style.set('stack')
            self.bfit.plt.figure(figstyle)
            draw_lines()
            self.bfit.draw_style.set('new')
        else:
            s = "Draw style not recognized"
            messagebox.showerror(message=s)
            self.logger.error(s)
            raise ValueError(s)
    
    # ======================================================================= #
    def export(self):
        """Export all data files as csv"""
        
        # filename
        filename = self.bfit.fileviewer.default_export_filename
        if not filename: return
        self.logger.info('Exporting to file %s',filename)
        try:
            filename = filedialog.askdirectory()+'/'+filename
        except TypeError:
            pass
        
        # get data and write
        for k in self.bfit.data.keys():
            d = self.bfit.data[k].bd
            self.bfit.export(d,filename%(d.year,d.run))
        self.logger.debug('Success.')
        
    # ======================================================================= #
    def get_data(self):
        """Split data into parts, and assign to dictionary."""
        
        self.logger.debug('Fetching runs')
    
        # make list of run numbers, replace possible deliminators
        try:
            run_numbers = self.string2run(self.run.get())
        except ValueError:
            self.logger.exception('Bad run number string')
            return
            
        # get the selected year
        year = int(self.year.get())
        
        # get data
        data = {}
        s = ['Failed to open run']
        for r in run_numbers:
            
            # get key for data storage
            runkey = self.bfit.get_run_key(r=r,y=year)
            
            # read from archive
            try:
                new_data = bdata(r,year=year)
            except (RuntimeError,ValueError):
                s.append("%d (%d)" % (r,year))
            else:
                
                # update data
                if runkey in self.bfit.data.keys():
                    self.bfit.data[runkey].bd = new_data
                    
                # new data
                else:
                    data[runkey] = fitdata(self.bfit,new_data)
            
        # print error message
        if len(s)>1:
            s = '\n'.join(s)
            print(s)
            self.logger.warning(s)
            messagebox.showinfo(message=s)
        
        # check that data is all the same runtype
        run_types = [self.bfit.data[k].mode for k in self.bfit.data.keys()]
        run_types = run_types + [data[k].mode for k in data.keys()]
        
        # different run types: select all runs of same type
        if not all([r==run_types[0] for r in run_types]):
            
            # unique run modes
            run_type_unique = np.unique(run_types)
            
            # message
            message = "Multiple run types detected:\n("
            for m in run_type_unique: 
                message += m+', '
            message = message[:-2]
            message += ')\n\nSelecting ' + run_types[0] + ' runs.'
            messagebox.showinfo(message=message)
            
        # get only run_types[0]
        self.logger.debug('Fetching runs of mode %s',run_types[0])
        for k in tuple(data.keys()):
            if data[k].mode == run_types[0]:
                self.bfit.data[k] = data[k]
            else:
                del data[k]
        
        try:
            self.runmode = run_types[0]
        except IndexError:
            s = 'No valid runs detected.'
            messagebox.showerror(message=s)
            self.logger.warning(s)
            raise RuntimeError(s)
        self.runmode_label['text'] = self.runmode_relabel[self.runmode]
        self.bfit.set_asym_calc_mode_box(self.runmode,self)
        self.bfit.set_asym_calc_mode_box(self.runmode,self.bfit.fit_files)
        
        keys_list = list(self.bfit.data.keys())
        keys_list.sort()
        
        # make lines
        n = 1
        for r in keys_list:
            
            # new line
            if r not in self.data_lines.keys():
                
                if r in self.data_lines_old.keys():
                    self.data_lines[r] = self.data_lines_old[r]
                    self.data_lines[r].set_label()
                    del self.data_lines_old[r]
                else:
                    self.data_lines[r] = dataline(\
                                            bfit = self.bfit,\
                                            lines_list = self.data_lines,\
                                            lines_list_old = self.data_lines_old,
                                            fetch_tab_frame = self.dataline_frame,\
                                            bdfit = self.bfit.data[r],\
                                            row = n)
            self.data_lines[r].grid(n)
            n+=1
            
        # remove old runs, modes not selected
        for r in tuple(self.data_lines.keys()):
            if self.data_lines[r].bdfit.mode != self.runmode:
                self.data_lines[r].degrid()
            
        self.logger.info('Fetched runs %s',list(data.keys()))
        
    # ======================================================================= #
    def remove_all(self):
        """Remove all data files from self.data_lines"""
        
        self.logger.info('Removing all data files')
        del_list = []
        for r in self.data_lines.keys():
            if self.data_lines[r].check_state.get():
                del_list.append(self.data_lines[r])
        for d in del_list:
            d.degrid()
    
    # ======================================================================= #
    def return_binder(self):
        """Switch between various functions of the enter button. """
        
        # check where the focus is
        focus_id = self.bfit.root.focus_get()
        
        # run or year entry
        if focus_id in [self.entry_run, self.entry_year]:
            self.logger.debug('Focus is: run or year entry')
            self.get_data()
        
        # checked rebin or checked run omission
        elif focus_id in [self.check_rebin_box,\
                          self.check_bin_remove_entry]:
            self.logger.debug('Focus is: checked rebin or checked run omission')
            self.set_all()
        elif focus_id == self.check_all_box:
            self.logger.debug('Focus is: check all box')
            self.draw_all()
        else:
            pass

    # ======================================================================= #
    def set_all(self):
        """Set a particular property for all checked items. """
        
        self.logger.info('Set all')
        
        # check all file lines
        for r in self.data_lines.keys():
            
            # if checked
            if self.data_lines[r].check_state.get():
                
                # get values to enter
                self.data_lines[r].rebin.set(self.check_rebin.get())
                new_text = self.check_bin_remove.get()
                
                # check for greyed text
                if new_text != self.bin_remove_starter_line:
                    self.data_lines[r].bin_remove.set(new_text)
                else:
                    self.data_lines[r].bin_remove.set("")
                    
                # generate focus out event: trigger grey text reset
                self.data_lines[r].bin_remove_entry.event_generate('<FocusOut>')

    # ======================================================================= #
    def set_all_labels(self):
        """Set lable text in all items """
        
        self.logger.info('Setting all label text')
        
        # check all file lines
        for r in self.data_lines.keys():
            self.data_lines[r].set_label()
            
        for r in self.data_lines_old.keys():
            self.data_lines_old[r].set_label()

    # ======================================================================= #
    def string2run(self,string):
        """Parse string, return list of run numbers"""
        
        # standardize deliminators
        full_string = string.replace(',',' ').replace(';',' ')
        full_string = full_string.replace(':','-')
        part_string = full_string.split()
        
        # get list of run numbers
        run_numbers = []
        for s in part_string:
            
            # select a range
            if '-' in s:
                
                # get runs in range
                try:
                    spl = s.split('-')
                    
                    # look for "range to current run"
                    if spl[1] == '':
                        
                        # look for latest run by run number
                        if int(spl[0]) < 45000:
                            dirloc = os.environ[self.bfit.bnmr_archive_label]
                        else:
                            dirloc = os.environ[self.bfit.bnqr_archive_label]
                            
                        runlist = glob.glob(os.path.join(dirloc,
                                                         str(self.year.get()),
                                                         '*.msr'))
                        spl[1] = max([int(os.path.splitext(os.path.basename(r))[0]) 
                                   for r in runlist])
                        
                    rn_lims = tuple(map(int,spl))
                    
                # get bad range first value
                except ValueError:
                    run_numbers.append(int(s.replace('-','')))
                
                # get range of runs
                else:
                    rns = np.arange(rn_lims[0],rn_lims[1]+1).tolist()
                    run_numbers.extend(rns)
            
            # get single run
            else:
                run_numbers.append(int(s))
        
        # sort
        run_numbers.sort()
        self.logger.debug('Parsed "%s" to run numbers (len: %d) %s',string,
                          len(run_numbers),run_numbers)
        
        if len(run_numbers) > self.max_number_fetched:
            raise RuntimeWarning("Too many files selected (max 50).")
        return run_numbers
    
    # ======================================================================= #
    def toggle_all(self):
        """Toggle all tickboxes"""
        self.logger.info('Toggling all tickboxes')
        for k in self.data_lines.keys():
            state = not self.data_lines[k].check_state.get()
            self.data_lines[k].check_state.set(state)

# =========================================================================== #
# =========================================================================== #
class dataline(object):
    """
        A line of objects to display run properties and remove bins and whatnot.
        
        bdfit:          fitdata object 
        bfit:           pointer to root 
        bin_remove:     StringVar for specifying which bins to remove in 1f runs
        bin_remove_entry: Entry object for bin remove 
        check:          Checkbox for selection (related to check_state)
        check_data:     BooleanVar for specifying to draw data
        check_fit:      BooleanVar for specifying to draw fit
        check_res:      BooleanVar for specifying to draw residual
        check_state:    BooleanVar for specifying check state
        draw_fit_checkbox: Checkbutton linked to check_fit
        draw_res_checkbox: Checkbutton linked to check_res
        id:             Str key for unique idenfication
        label:          StringVar for labelling runs in legends
        label_entry:    Entry object for labelling runs in legends
        line_frame:     Frame that object is placed in
        lines_list:     dictionary of datalines
        lines_list_old: dictionary of datalines
        mode:           bdata run mode
        rebin:          IntVar for SLR rebin
        row:            position in list
        run:            bdata run number
        year:           bdata year
    """
        
    bin_remove_starter_line = '24 100-200 (bins)'
    
    # ======================================================================= #
    def __init__(self,bfit,lines_list,lines_list_old,fetch_tab_frame,bdfit,row):
        """
            Inputs:
                fetch_tab_frame: parent in which to place line
                bdfit: fitdata object corresponding to the file which is placed here. 
                row: where to grid this object
        """
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing run %d (%d)',bdfit.run,bdfit.year)
        
        # variables
        self.bfit = bfit
        self.bin_remove = bdfit.omit
        self.label = bdfit.label
        self.rebin = bdfit.rebin
        self.check_state = bdfit.check_state
        self.mode = bdfit.mode
        self.run =  bdfit.run
        self.year = bdfit.year
        self.row = row
        self.lines_list = lines_list
        self.lines_list_old = lines_list_old
        bd = bdfit.bd
        self.bdfit = bdfit
        self.id = bdfit.id
        
        # build objects
        line_frame = Frame(fetch_tab_frame)
        line_frame.bind('<Enter>', self.on_line_enter)
        line_frame.bind('<Leave>', self.on_line_leave)
        
        bin_remove_entry = ttk.Entry(line_frame,textvariable=self.bin_remove,\
                width=17)
                
        label_label = ttk.Label(line_frame,text="Label:",pad=5)
        self.label_entry = ttk.Entry(line_frame,textvariable=self.label,\
                width=18)
                
        remove_button = ttk.Button(line_frame,text='Remove',\
                command=self.degrid,pad=1)
        draw_button = ttk.Button(line_frame,text='Draw',
                                 command=lambda:self.draw(figstyle='data'),
                                 pad=1)
        
        self.check_data = BooleanVar()
        self.check_data.set(True)
        draw_data_checkbox = ttk.Checkbutton(line_frame,text='Data',
                variable=self.check_data,onvalue=True,offvalue=False,pad=5)
        
        self.check_fit = BooleanVar()
        self.check_fit.set(False)
        self.draw_fit_checkbox = ttk.Checkbutton(line_frame,text='Fit',
                variable=self.check_fit,onvalue=True,offvalue=False,pad=5,
                state=DISABLED)
        
        self.check_res = BooleanVar()
        self.check_res.set(False)
        self.draw_res_checkbox = ttk.Checkbutton(line_frame,text='Res',
                variable=self.check_res,onvalue=True,offvalue=False,pad=5,
                state=DISABLED)
        
        rebin_label = ttk.Label(line_frame,text="Rebin:",pad=5)
        rebin_box = Spinbox(line_frame,from_=1,to=100,width=3,\
                textvariable=self.rebin)
        self.rebin.set(self.bfit.fetch_files.check_rebin.get())
                
        self.check_state.set(bfit.fetch_files.check_state.get())
        self.check = ttk.Checkbutton(line_frame,variable=self.check_state,\
                onvalue=True,offvalue=False,pad=5)
         
        self.set_check_text()
         
        # add grey text to bin removal
        bin_remove_entry.insert(0,self.bin_remove_starter_line)
        entry_fn = partial(on_entry_click,\
                text=self.bin_remove_starter_line,entry=bin_remove_entry)
        on_focusout_fn = partial(on_focusout,\
                text=self.bin_remove_starter_line,entry=bin_remove_entry)
        bin_remove_entry.bind('<FocusIn>', entry_fn)
        bin_remove_entry.bind('<FocusOut>', on_focusout_fn)
        bin_remove_entry.config(foreground=colors.entry_grey)
             
        # add grey text to label
        self.set_label()
                
        # grid
        c = 1
        self.check.grid(column=c,row=0,sticky=E); c+=1
        if self.mode in ['1f','1n','1w']: 
            bin_remove_entry.grid(column=c,row=0,sticky=E); c+=1
        if self.mode in ['20','2h']: 
            rebin_label.grid(column=c,row=0,sticky=E); c+=1
            rebin_box.grid(column=c,row=0,sticky=E); c+=1
        label_label.grid(column=c,row=0,sticky=E); c+=1
        self.label_entry.grid(column=c,row=0,sticky=E); c+=1
        draw_data_checkbox.grid(column=c,row=0,sticky=E); c+=1
        self.draw_fit_checkbox.grid(column=c,row=0,sticky=E); c+=1
        self.draw_res_checkbox.grid(column=c,row=0,sticky=E); c+=1
        draw_button.grid(column=c,row=0,sticky=E); c+=1
        remove_button.grid(column=c,row=0,sticky=E); c+=1
        
        # resizing
        fetch_tab_frame.grid_columnconfigure(0, weight=1)   # big frame
        for i in (3,5,7):
            line_frame.grid_columnconfigure(i, weight=100)    # input labels
        for i in (4,6,8):
            line_frame.grid_columnconfigure(i, weight=1)  # input fields
        
        # passing
        self.line_frame = line_frame
        self.bin_remove_entry = bin_remove_entry
        
    # ======================================================================= #
    def __del__(self):
        
        # kill buttons and frame
        try:
            for child in self.line_frame.winfo_children():
                child.destroy()
        except Exception:
            pass
            
    # ======================================================================= #
    def grid(self,row):
        """Re-grid a dataline object so that it is in order by run number"""
        self.row = row
        self.line_frame.grid(column=0,row=row,columnspan=2, sticky=(W,N))
        self.line_frame.update_idletasks()
        self.bfit.data[self.id] = self.bdfit
        self.set_check_text()
        
    # ======================================================================= #
    def degrid(self):
        """Hide displayed dataline object from file selection. """
        
        self.logger.info('Degridding run %d (%d)',self.run,self.year)
        
        self.line_frame.grid_forget()
        self.line_frame.update_idletasks()
        
        self.lines_list_old[self.id] = self.lines_list[self.id]
        del self.lines_list[self.id]
        del self.bfit.data[self.id]
        
        # repopulate fit files tab
        self.bfit.fit_files.populate()
        
        # remove data from storage
        if len(self.lines_list) == 0:
            ff = self.bfit.fetch_files
            ff.runmode_label['text'] = ''
                
    # ======================================================================= #
    def draw(self,figstyle):
        """
            Draw single data file.
            
            figstyle: one of "data", "fit", or "param" to choose which figure 
                    to draw in
        """
        
        # draw data
        if self.check_data.get():
            self.logger.debug('Draw run %d (%d)',self.run,self.year)
                    
            # get new data file
            data = self.bfit.data[self.bfit.get_run_key(r=self.run,y=self.year)]
            data.read()
            
            # get data file run type
            d = self.bfit.fetch_files.asym_type.get()
            d = self.bfit.asym_dict[d]
            
            if self.bin_remove.get() == self.bin_remove_starter_line:
                self.bfit.draw(data,d,self.rebin.get(),figstyle=figstyle,
                    label=self.label.get())
            else:
                self.bfit.draw(data,d,self.rebin.get(),figstyle=figstyle,\
                    option=self.bin_remove.get(),label=self.label.get())

        # draw fit
        if self.check_fit.get():
            if self.check_data.get():
                mode = self.bfit.draw_style.get()
                self.bfit.draw_style.set('stack')
                self.bfit.fit_files.draw_fit(id=self.id,figstyle=figstyle)
                self.bfit.draw_style.set(mode)
            else:
                self.bfit.fit_files.draw_fit(id=self.id,figstyle=figstyle)
                
        # draw residual
        if self.check_res.get():
            self.bfit.fit_files.draw_residual(id=self.id,
                                              figstyle=figstyle,
                                              rebin=self.rebin.get())

    # ======================================================================= #
    def set_check_text(self):
        """Update the string for the check state box"""
        
        bdfit = self.bdfit
        
        # temperature
        try:
            self.temperature = int(np.round(bdfit.temperature.mean))
        except AttributeError:
            self.temperature = -999
            
        # field (Tesla)
        if bdfit.field > 0.1:
            self.field = np.around(bdfit.field,2)
            
            try:
                field_text = "%3.2fT"%self.field
            except TypeError:
                field_text = ' '
        else:
            self.field = np.round(bdfit.field*1e4)
            
            try:
                field_text = "%3.0fG"%self.field
            except TypeError:
                field_text = ' '
        
        # bias
        self.bias = self.bdfit.bias
        try:
            bias_text = "%4.2fkV"%self.bias
        except TypeError:
            bias_text = ' '
        
        # duration 
        try: 
            duration_text = "%02d:%02d" % divmod(self.bdfit.duration, 60)
        except AttributeError:
            duration_text = ' '
        
        # set the text    
        info_str = "%d.%d, %3dK, %s, %s, %s" % (self.year,self.run,
                                              self.temperature,field_text,
                                              bias_text,duration_text)
        self.check.config(text=info_str)
    
    # ======================================================================= #
    def set_label(self):
        """Set default label text"""
        
        # get default label
        try:
            label = self.bfit.get_label(self.bfit.data[self.id])
        except KeyError:
            return
        
        # clear old text
        self.label_entry.delete(0,'end')
        
        # set
        self.label_entry.insert(0,label)
        
        # make function to clear and replace with default text
        entry_fn_lab = partial(on_entry_click,text=label,
                               entry=self.label_entry)
        on_focusout_fn_lab = partial(on_focusout,text=label,
                                 entry=self.label_entry)
                                 
        # bindings
        self.label_entry.bind('<FocusIn>', entry_fn_lab)
        self.label_entry.bind('<FocusOut>', on_focusout_fn_lab)
        self.label_entry.config(foreground=colors.entry_grey)
        
    # ======================================================================= #
    def on_line_enter(self,*args):
        """Make the dataline grey on mouseover"""
        self.line_frame.config(bg=colors.focusbackground)
    
    # ======================================================================= #
    def on_line_leave(self,*args):
        """Make the dataline black on stop mouseover"""
        self.line_frame.config(bg=colors.background)



# Fit list of bdata objects with function list
# Derek Fujimoto
# Nov 2018

import collections
import numpy as np
from bdata import bdata
from scipy.optimize import curve_fit
from tqdm import tqdm
from bfit.fitting.global_bdata_fitter import global_bdata_fitter

# ========================================================================== #
def fit_list(runs,years,fnlist,omit=None,rebin=None,sharelist=None,npar=-1,
              hist_select='',xlims=None,asym_mode='c',fixed=None,**kwargs):
    """
        Fit combined asymetry from bdata.
    
        runs:           list of run numbers
        
        years:          list of years corresponding to run numbers, or int which applies to all
        
        fnlist:         list of function handles to fit (or single which applies to all)
                        must specify inputs explicitly (do not do def fn(*par)!)
                        must have len(fn) = len(runs) if list
        
        omit:           list of strings of space-separated bin ranges to omit
        rebin:          list of rebinning of data prior to fitting. 
        
        sharelist:      list of bool to indicate which parameters are shared. 
                        True if shared
                        len = number of parameters.
        
        npar:           number of free parameters in each fitting function.
                        Set if number of parameters is not intuitable from 
                            function code.      
        
        hist_select:    string for selecting histograms to use in asym calc
        
        xlims:          list of 2-tuple for (low,high) bounds on fitting range 
                            based on x values
        
        asym_mode:      input for asymmetry calculation type 
                            c: combined helicity
                            h: split helicity
                            
                        For 2e mode, prefix with:
                            sl_: combined timebins using slopes
                            dif_: combined timebins using differences
                            raw_: raw time-resolved 
                            
                            ex: sl_c or raw_h or dif_c
        
        fixed:          list of booleans indicating if the paramter is to be 
                        fixed to p0 value (same length as p0). Returns best 
                        parameters in order presented, with the fixed 
                        parameters omitted. Can be a list of lists with one list 
                        for each run.
        
        kwargs:         keyword arguments for curve_fit. See curve_fit docs. 
        
        Returns: (par,cov,ch,gchi)
            par: best fit parameters
            cov: covariance matrix
            chi: chisquared of fits
            gchi:global chisquared of fits
    """
    
    nruns = len(runs)
    
    # get fnlist
    if not isinstance(fnlist,collections.Iterable):
        fnlist = [fnlist]
    
    # get number of parameters
    if npar < 0:
        npar = fnlist[0].__code__.co_argcount-1

    # get fnlist again
    fnlist.extend([fnlist[-1] for i in range(nruns-len(fnlist))])

    # get sharelist
    if sharelist is None:
        sharelist = np.zeros(npar,dtype=bool)

    # get omit
    if omit is None:
        omit = ['']*nruns
    elif len(omit) < nruns:
        omit = np.concatenate(omit,['']*(nruns-len(omit)))
        
    # get rebin
    if rebin is None:
        rebin = np.ones(nruns)
    elif type(rebin) is int:
        rebin = np.ones(nruns)*rebin
    elif len(rebin) < nruns:
        rebin = np.concatenate((rebin,np.ones(nruns-len(rebin))))
    
    rebin = np.asarray(rebin).astype(int)
    
    # get years
    if type(years) in (int,float):
        years = np.ones(nruns,dtype=int)*years
        
    # get p0 list
    if 'p0' in kwargs.keys():
        p0 = kwargs['p0']
        del kwargs['p0']
    else:
        p0 = [np.ones(npar)]*nruns

    # fit globally -----------------------------------------------------------
    if any(sharelist) and len(runs)>1:
        print('Running shared parameter fitting...')
        g = global_bdata_fitter(runs,years,fnlist,sharelist,npar,xlims,
                                asym_mode=asym_mode,rebin=rebin,fixed=fixed)
        g.fit(p0=p0,**kwargs)
        gchi,chis = g.get_chi() # returns global chi, individual chi
        pars,covs = g.get_par()
        
    # fit runs individually --------------------------------------------------
    else:
        
        # get bounds
        if 'bounds' in kwargs.keys():
            bounds = kwargs['bounds']
            del kwargs['bounds']
        else:
            bounds = [(-np.inf,np.inf)]*nruns 
            
        # check p0 dimensionality
        if len(np.asarray(p0).shape) < 2:
            p0 = [p0]*nruns
        
        # check xlims shape - should match number of runs
        if xlims is None:
            xlims = [None]*nruns
        elif len(np.asarray(xlims).shape) < 2:
            xlims = [xlims for i in range(nruns)]
        else:
            xlims = list(xlims)
            xlims.extend([xlims[-1] for i in range(len(runs)-len(xlims))])
        
        # check fixed shape
        if fixed is not None: 
            fixed = np.asarray(fixed)
            if len(fixed.shape) < 2:
                fixed = [fixed]*nruns
        else:
            fixed = [[False]*npar]*nruns
        
        pars = []
        covs = []
        chis = []
        gchi = 0.
        dof = 0.
        
        iter_obj = tqdm(zip(runs,years,fnlist,omit,rebin,p0,bounds,xlims,fixed),
                        total=len(runs),desc='Independent Fitting')
        for r,yr,fn,om,re,p,b,xl,fix in iter_obj:
            
            # get data for chisq calculations
            x,y,dy = _get_asym(bdata(r,year=yr),asym_mode,rebin=re,omit=om)
            
            # get x limits
            if xl is None:  
                xl = [-np.inf,np.inf]
            else:
                if xl[0] is None: xl[0] = -np.inf
                if xl[1] is None: xl[1] = np.inf
            
            # get good data
            idx = (xl[0]<x)*(x<xl[1])*(dy!=0)
            x = x[idx]
            y = y[idx]
            dy = dy[idx]
            
            # trivial case: all parameters fixed
            if all(fix):
                lenp = len(p)
                s = np.full((lenp,lenp),np.nan)
                c = np.sum(np.square((y-fn(x,*p))/dy))/len(y)
                
            # fit with free parameters
            else:            
                p,s,c = fit_single(r,yr,fn,om,re,hist_select,p0=p,bounds=b,
                                   xlim=xl,asym_mode=asym_mode,fixed=fix,**kwargs)
            # outputs
            pars.append(p)
            covs.append(s)
            chis.append(c)
            
            # get global chi             
            gchi += np.sum(np.square((y-fn(x,*p))/dy))
            dof += len(x)-len(p)
        gchi /= dof
        
    pars = np.asarray(pars)
    covs = np.asarray(covs)
    chis = np.asarray(chis)
            
    return(pars,covs,chis,gchi)

# =========================================================================== #
def fit_single(run,year,fn,omit='',rebin=1,hist_select='',xlim=None,asym_mode='c',
               fixed=None,**kwargs):
    """
        Fit combined asymetry from bdata.
    
        runs:           run number
        
        years:          year
        
        fn:             function handle to fit
        
        omit:           string of space-separated bin ranges to omit
        rebin:          rebinning of data prior to fitting. 
        
        hist_select:    string for selecting histograms to use in asym calc
        
        xlim:           2-tuple for (low,high) bounds on fitting range based on 
                            x values
        
        asym_mode:      input for asymmetry calculation type 
                            c: combined helicity
                            h: split helicity
                            
                        For 2e mode, prefix with:
                            sl_: combined timebins using slopes
                            dif_: combined timebins using differences
                            raw_: raw time-resolved
                            
                            ex: sl_c or raw_h or dif_c
        
        fixed:          list of booleans indicating if the paramter is to be 
                        fixed to p0 value (same length as p0). Returns best 
                        parameters in order presented, with the fixed 
                        parameters omitted.
        
        kwargs:         keyword arguments for curve_fit. See curve_fit docs. 
        
        Returns: (par,cov,chi)
            par: best fit parameters
            cov: covariance matrix
            chi: chisquared of fit
    """
    
    # Get data input
    data = bdata(run,year)
    x,y,dy = _get_asym(data,asym_mode,rebin=rebin,omit=omit)
            
    # check for values with error == 0. Omit these values. 
    tag = dy != 0
    x = x[tag]
    y = y[tag]
    dy = dy[tag]
    
    # apply xlimits
    if xlim is not None:
        tag =(xlim[0]<x)*(x<xlim[1])
        x = x[tag]
        y = y[tag]
        dy = dy[tag]
    
    # p0
    if 'p0' not in kwargs:
        kwargs['p0'] = np.ones(fn.__code__.co_argcount-1)
    
    # fixed parameters
    did_fixed = False
    if fixed is not None and any(fixed):
        
        # save stuff for inflation
        did_fixed = True
        p0 = np.copy(kwargs['p0'])
        npar = len(p0)
        
        # dumb case: all values fixed: 
        if all(fixed):
            cov = np.full((npar,npar),np.nan)
            chi = np.sum(np.square((y-fn(x,*p0))/dy))/len(y)
            return (p0,cov,chi)
        
        # prep inputs
        fixed = np.asarray(fixed)
        if 'bounds' in kwargs:  bounds = kwargs['bounds']
        else:                   bounds = None
        
        # get fixed version
        fn,kwargs['p0'],bounds = _get_fixed_values(fixed,fn,kwargs['p0'],bounds)
        
        # modify fiting inputs
        if bounds is not None:  kwargs['bounds'] = bounds
        
    # Fit the function 
    par,cov = curve_fit(fn,x,y,sigma=dy,absolute_sigma=True,**kwargs)
    dof = len(y)-len(kwargs['p0'])
    
    # get chisquared
    chi = np.sum(np.square((y-fn(x,*par))/dy))/dof
    
    # inflate parameters with fixed values 
    if did_fixed:
        
        # inflate parameters
        par_inflated = np.zeros(npar)
        par_inflated[fixed] = p0[fixed]
        par_inflated[~fixed] = par
        par = par_inflated
        
        # inflate cov matrix with NaN
        nfixed_flat = np.concatenate(np.outer(~fixed,~fixed))
        c_inflated = np.full(npar**2,np.nan)
        c_inflated[nfixed_flat] = np.concatenate(cov)
        cov = c_inflated.reshape(npar,-1)
    
    return (par,cov,chi)
    
# =========================================================================== #
def _get_asym(data,asym_mode,**asym_kwargs):
    """
        Get asymmetry
        
        data = bdata object
        asym_mode: mode as described above
    """
    
    if asym_mode in ('c','p','n','sc','dc','sl_c','dif_c'):
        x,y,dy = data.asym(asym_mode,**asym_kwargs)
    elif asym_mode in ('h','sh','dh','sl_h','dif_h'):
        raise RuntimeError('Split helicity fitting not yet implemented')
    elif 'raw' in asym_mode:
        raise RuntimeError('2e Time-resolved fitting not yet implemented')

    return (x,y,dy)
    
# =========================================================================== #
def _get_fixed_values(fixed,fn,p0,bounds=None):
    """
        Get fixed function, p0, bounds
    """
    
    # save original inputs
    fn_orig = fn
    p0_orig = np.copy(p0)
    npar_orig = len(p0_orig)
            
    # index of fixed parameters
    idx = np.where(fixed)
    
    # make new fitting function with fixed parameter(s)
    def fn(x,*args):
        args_fixed = np.zeros(npar_orig)
        args_fixed[fixed] = p0_orig[fixed]
        args_fixed[~fixed] = args
        return fn_orig(x,*args_fixed)
    
    # make new p0
    p0 = np.asarray(p0_orig)[~fixed]
    
    # bounds
    if bounds is not None:
        try:
            bounds[0] = np.asarray(bounds[0])[~fixed]
            bounds[1] = np.asarray(bounds[1])[~fixed]
        except IndexError:
            pass
    
    return (fn,p0,bounds)



__all__=['functions',
         'global_fitter',
         'global_bdata_fitter',
         'PulsedFns',
         'pulsed',
         'continuous',
         'fitter',
         'fit_bdata',
         'decay_31mg']
